(window.webpackJsonp=window.webpackJsonp||[]).push([[37,86,157],{782:function(o,t,e){"use strict";e.r(t);var n={props:{url:{type:String},icon:{type:String},iconColor:{type:String,default:function(){return"white"}},tooltip:{type:String},showBadge:{type:Boolean,default:!1},badgeVerticalPosition:{type:String},badgeHorizontalPosition:{type:String},badgeBorderColor:{type:String}},methods:{onClickIcon:function(){this.$emit("click-icon")}}},r=e(31),component=Object(r.a)(n,(function(){var o=this,t=o._self._c;return t("BaseButton",{staticClass:"icon-with-badge",attrs:{"data-title":o.tooltip},on:{"on-click":o.onClickIcon}},[t("i",{directives:[{name:"badge",rawName:"v-badge",value:{showBadge:o.showBadge,verticalPosition:o.badgeVerticalPosition,horizontalPosition:o.badgeHorizontalPosition,borderColor:o.badgeBorderColor},expression:"{\n      showBadge: showBadge,\n      verticalPosition: badgeVerticalPosition,\n      horizontalPosition: badgeHorizontalPosition,\n      borderColor: badgeBorderColor,\n    }"}],key:o.showBadge,staticClass:"icon-wrapper"},[t("svgicon",{attrs:{name:o.icon,width:"22",height:"22",color:o.iconColor}})],1)])}),[],!1,null,null,null);t.default=component.exports;installComponents(component,{BaseButton:e(460).default})}}]);