(window.webpackJsonp=window.webpackJsonp||[]).push([[152],{1196:function(e,t,n){"use strict";n.r(t);var o={props:{options:{type:Array,required:!0},selectedOption:{type:String}},model:{prop:"selectedOption",event:"change"},methods:{onChangeOption:function(option){this.$emit("change",option)}}},l=n(31),component=Object(l.a)(o,(function(){var e=this,t=e._self._c;return e.options.length?t("div",[t("RadioButtonsSelectBase",{attrs:{options:e.options,"selected-option":e.selectedOption},on:{change:e.onChangeOption}})],1):e._e()}),[],!1,null,null,null);t.default=component.exports;installComponents(component,{RadioButtonsSelectBase:n(1186).default})}}]);