(window.webpackJsonp=window.webpackJsonp||[]).push([[216],{1289:function(n,e){n.exports={attributes:{title:"spaCy",description:"The ArgillaSpacyTrainer leverages the features of spaCy to train programmatically with Argilla.",links:[{linkText:"Argilla docs",linkLink:"https://docs.argilla.io/en/latest/guides/train_a_model.html"},{linkText:"spaCy docs",linkLink:"https://spacy.io/usage/training"}]},html:'<p><em>code snippet</em></p>\n<pre><code class="language-python">from argilla.training import ArgillaTrainer\n\ntrainer = ArgillaTrainer(\n    name=&quot;&lt;my_dataset_name&gt;&quot;,\n    workspace=&quot;&lt;my_workspace_name&gt;&quot;,\n    framework=&quot;spacy&quot;,\n    train_size=0.8\n)\ntrainer.update_config(max_epochs=10)\ntrainer.train(output_dir=&quot;token-classification&quot;)\nrecords = trainer.predict(&quot;The ArgillaTrainer is great!&quot;, as_argilla_records=True)\n</code></pre>\n<p><em>update training config</em></p>\n<pre><code class="language-python"># `spacy.training`\ntrainer.update_config(\n    dev_corpus = &quot;corpora.dev&quot;,\n    train_corpus = &quot;corpora.train&quot;,\n    seed = 42,\n    gpu_allocator = 0,\n    accumulate_gradient = 1,\n    patience = 1600,\n    max_epochs = 0,\n    max_steps = 20000,\n    eval_frequency = 200,\n    frozen_components = [],\n    annotating_components = [],\n    before_to_disk = None,\n    before_update = None\n)\n</code></pre>\n'}}}]);